#!/usr/bin/env python

"""
CourseBuilder

Coursebuilder is a preprocessor tool for [pandoc](https://pandoc.org)
to generate multi-lingual curricula documentation tables from
structured representations as a flatfile database. Data scheme and
actual values are kept in YAML files in order to version them with git.

"""

from argparse import ArgumentParser
import os,sys
import yaml
import pandas as pd
from string import Template

from tablegenerator import TableGenerator
from markdowngenerator import MarkdownGenerator
from templategenerator import TemplateGenerator
from schema import Schema


class CourseBuilder:

    @staticmethod
    def generate(args):
        if args.schema and args.meta:

            # get actual fields
            actual_fields = None

            # use a file instead of list
            if args.fields and os.path.isfile(args.fields[0]):
                with open(args.fields[0]) as ff:
                    actual_fields = yaml.load(ff,Loader=yaml.Loader)['fields']
            else:
                # seem we have a list or None
                actual_fields = args.fields

            # get schema
            schema = None
            with open(args.schema) as f:
                schema = Schema(yaml.load(f,Loader=yaml.Loader))

            # if no fields are given, take all!
            if actual_fields == None:
                actual_fields = list(schema.keys())


            result_df = []

            # iterate through meta files
            for m in args.meta:
                with open(m) as fm:

                    if args.legacy:

                        MarkdownGenerator.generate_table_legacy(
                            table_items=schema.to_list_of_tuple(
                                meta=yaml.load(fm,Loader=yaml.Loader),
                                fields=actual_fields,
                                lang=args.lang),
                            add_pagebreak=args.pagebreak,
                            title_template=args.title,
                            first_colwidth=args.leftcol)
                    elif args.query:

                        lot = schema.to_short_dict(
                                meta=yaml.load(fm,Loader=yaml.Loader),
                                fields=actual_fields,
                                lang=args.lang)

                        result_df.append(pd.DataFrame([lot]))
                    else:
                            MarkdownGenerator.generate_table(
                            table_items=schema.to_list_of_tuple(
                                meta=yaml.load(fm,Loader=yaml.Loader),
                                fields=actual_fields,
                                lang=args.lang),
                            add_pagebreak=args.pagebreak,
                            title_template=args.title,
                            first_colwidth=args.leftcol)

            # query mode
            if args.query and len(result_df):

                # got the list
                df = pd.concat(result_df,ignore_index=True)

                # generate a dataframe
                df_q = df.query(args.query)

                # generate a compound column --query-compound column:sum
                if args.query_compound:
                    print(args.query_compound)
                    df_q.loc[:,'form-of-instruction.sum'] = df_q['form-of-instruction'].apply(lambda x: sum(list(x.values())))

                # --query-sort is parameterized as min:credits - hence direction:column
                if args.query_sort:
                    qs = args.query_sort.split(':')
                    match qs[0]:
                        case 'min' : df_q = df_q.sort_values(by=qs[1],ascending=True,key=lambda col: min(col) if hasattr(col,'__len()__') else col)
                        case 'max' : df_q = df_q.sort_values(by=qs[1],ascending=False,key=lambda col: max(col) if hasattr(col,'__len()__') else col)

                # filter query
                if args.query_filter:
                    df_q = df_q.loc[:,args.query_filter]

                # print(df_q.head())

                # set value transforms
                if args.query_template:

                    # no idea yet how to parameterize this
                    ww = 'written'
                    #df_q['form-of-exam'] = 'Schriftlich' if df_q.loc[:,'form-of-exam'] == 'written' else 'was anderes'
                    # mm = Template("{'written':'S','oral':'mündlich'}[${v}]")?
                    # print(mm.format(v=mm))

                # lets get crazy to create a summary table!
                df_summary = pd.DataFrame([{
                    'sum.credits': df_q['credits'].sum()
                }])

                # set labels directly!
                if args.query_labels:
                    df_q.columns = args.query_labels

                q_as_md = df_q.to_markdown(tablefmt='grid',index=False)

                print(q_as_md)

                print(df_summary.to_markdown(tablefmt='grid',index=False))


    @staticmethod
    def run():

        # arguments
        parser = ArgumentParser(description='versatile curricula generator')

        # loading mode for internal database
        parser.add_argument('-m','--meta',action="extend", nargs="+", type=str,help="course description(s) as YAML file(s)")
        parser.add_argument('-l','--lang',help="Language to parse from meta file (use de or en)",default='de')
        parser.add_argument('-f','--fields',help="Fields to be used, the table will be build accordingly",action="extend", nargs="+", type=str)
        parser.add_argument('-s','--schema', help="using provided schema")

        # query mode
        parser.add_argument('-q','--query', type=str, default=None, help="compound query to select items")
        parser.add_argument('-qs','--query-sort',type=str,default=None,help="sort query with a min/max over a column like min:credits")
        parser.add_argument('-qc','--query-compound',type=str,default=None,help="create a compound from a column with multiple values/dictionaries in cells")
        parser.add_argument('-qf','--query-filter',type=str,default=[],action="extend", nargs="+",help="filter final list of columns for output")
        parser.add_argument('-ql','--query-labels',type=str,default=[],action="extend", nargs="+",help="new labels for query like")
        parser.add_argument('-qt','--query-template',type=str,default=[],action="extend", nargs="+",help="templates for values in the form of {value}")


        # create pagebreaks
        parser.add_argument('-p','--pagebreak',action="store_true",help="add a pagebreak after each module")
        parser.add_argument('--title',type=str,default=None,help="template for title - use curly brackets (i.e. {}) to mark where the title string is inserted")
        parser.add_argument('-b','--book',type=str,help="process a whole curriculum book with sections")
        parser.add_argument('--level',type=int,default=1,help="level of header tags")
        parser.add_argument('--table-gen',type=str,default=None,help='runs table generator')
        parser.add_argument('--template',type=str,default=None,help='defines a template to be used with fields')
        parser.add_argument('-o','--out',type=str,default=None,help='set the output type')
        parser.add_argument('--legacy',action="store_true",help="use legacy generator mode for compatibility")

        parser.add_argument('--leftcol',type=int,default=35,help='maximum size of left column')

        # get arguments
        args = parser.parse_args()

        if args.table_gen:

            tg = TableGenerator()
            tg.generate_table(args.table_gen)

            return

        # book mode with predefined setting from a book file
        if args.book and args.schema:

            with open(args.book) as bf:

                actual_fields = []

                book = yaml.load(bf,Loader=yaml.Loader)
                book_path = os.path.abspath(args.book)

                for bi in book['book']:

                    if 'fields' in bi:
                        actual_fields = bi['fields']

                    if 'sections' in bi:
                        for section in bi['sections']:

                            if 'text' in section:
                                print(section['text'][args.lang])

                            # gernerate section wise parts
                            if 'modules' in section:

                                # override fields
                                args.fields = actual_fields

                                # expand filenames to be relative to the book
                                args.meta = [os.path.join(os.path.dirname(book_path),mod_path) for mod_path in section['modules']]

                                CourseBuilder.generate(args=args)

        # verbose command line mode
        elif args.schema:
            CourseBuilder.generate(args=args)
        else:
            parser.print_help()

# run as main
if __name__ == '__main__':
    # recommended setting for pandas
    pd.options.mode.copy_on_write = True
    # run
    CourseBuilder.run()
